import typer
from uvicore import app, config


cli = typer.Typer()

@cli.command()
def version(version: bool = typer.Option(None, "--version", "-v")):
    """Uvicorn Version"""
    typer.echo('1.0.0')

    #server = import_module(app.name + '.http.server.Web')[0]()
    #server.run()


    #app.create_flask_app()
    #app.serve_flask()


# class Flask(import_name: str,
# static_url_path: Optional[str]=...,
# static_folder: Optional[str]=...,
# static_host: Optional[str]=...,
# host_matching: bool=...,
# subdomain_matching: bool=...,
# template_folder: str=...,
# instance_path: Optional[str]=...,
# instance_relative_config: bool=...,
# root_path: Optional[str]=...)



import os
import typer
import uvicorn
from uvicore import config

from subprocess import Popen


cli = typer.Typer()

@cli.command()
def serve():
    """Unicorn dev server (reload and logs)
    """
    # Uvicorn dev server info
    appname = config('app.server.app')
    host = config('app.server.host')
    port = config('app.server.port')
    autoreload = config('app.server.reload')
    access_log = config('app.server.access_log')

    # Run Uvicorn server
    uvicorn.run(appname, host=host, port=port, reload=autoreload, access_log=access_log)



class Str:
    def __init__(self, data):
        self.__data = data

    def contains(self, value):
        return value in self.__data

    def append(self, value):
        return str(self.__data) + str(value)

    def upper(self):
        return str(self.__data).upper()

    def lower(self):
        return str(self.__data).lower()


class Obj:
    def __init__(self, data):
        self.__dict__ = data

    def dict(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith('__')}

    def __repr__(self):
        return str({k: v for k, v in self.__dict__.items() if not k.startswith('__')})


class Collection():
    # _names are protected
    # __names are private
    # @property for getters
    # @varname.setter for setters

    def __init__(self, data):
        #self.test = 'asdf'
        self.__i = 0
        self.__data = []
        #self.Object = lambda **kwargs: type("Object", (), kwargs)
        for item in data:
            #self.__data.append(self.__object(item))
            self.__data.append(Obj(item))

    def dict(self):
        items = []
        for data in self.__data:
            items.append(data.dict())
        return items

    def filter(self, callback):
        return filter(callback, self.__data)

    # def __object(self, dict):
    #     return type("Object", (), dict)

    def add(self, item):
        #self.__data.append(self.__object(item))
        self.__data.append(Obj(item))

    def __iter__(self):
        return iter(self.__data)

    def __getitem__(self, item):
        return self.__data[item]

    # @property
    # def test(self):
    #     return self.__test

    # @test.setter
    # def test(self, value):
    #     self.__test = value

    # def __next__(self):
    #     #return self.next()
    #     print('1')
    #     if len(self.data) <= self.i:
    #         self.i += 1
    #         #return self.data[self.i]
    #         return self.i
    #     else:
    #         raise StopIteration()




from typing import List
from uvicore import app, config
from uvicore.support.click import click, group_kargs, typer
from uvicore.support.helpers import import_module
from uvicore.foundation.application import Application

class ServiceProvider:

    app: Application

    def __init__(self, app):
        self.app = app

    def routes(self, prefix, router):
        # Don't load commands if not running in CLI
        if app.is_console: return

        (Router, method, path) = import_module(router)
        Router(app, prefix).register()


    def commands(self, *, name: str, help: str = None, commands: List, force: bool = False):
        # Don't load commands if not running in CLI
        if not force and not app.is_console: return

        # Defining the name as 'root' makes the commands a root level command
        # NOT a click subcommand nested under a name

        if name != 'root':
            # Create a new click group for all commands in this app
            @click.group(**group_kargs, help=help)
            def group():
                pass

        # Add each apps commands to their own group
        for command_name, module in commands:
            (mod, method, path) = import_module(module)
            if name == 'root':
                # Add all uvicore commands to main command (NOT an app based subcommand)
                app.cli.add_command(typer.main.get_command(mod), command_name)
            else:
                # Add all apps commands to a click subcommand
                group.add_command(typer.main.get_command(mod), command_name)

        if name != 'root':
            app.cli.add_command(group, name)


    def configs(self, modules: List):
        for name, module in modules:
            (mod, method, path) = import_module(module)
            config.set(name, mod)



import time
import os
import importlib


def import_module(module):
    """Import module from string"""
    parts = module.split('.')
    path = '.'.join(parts[0:-1])
    name = ''.join(parts[-1:])
    return (getattr(importlib.import_module(path), name), name, path)


def find_base_path(file):
    """
    Finds app base path relative to some __file__ in the app
    """
    count = 0
    path=os.path.dirname(os.path.realpath(file))
    while count < 50:
        if (
            os.path.exists(os.path.realpath(path + '/uvicorex')) or
            os.path.exists(os.path.realpath(path + '/setup.py')) or
            os.path.exists(os.path.realpath(path + '/.env'))
        ):
            return os.path.realpath(path)
        else:
            path = os.path.realpath(path + '/../')
        count += 1
    exit("Could not find base path")



# For some reason at this level, causes circular dependency
# ImportError: cannot import name 'config' from partially initialized module 'mrcore' (most likely due to a circular import)
#from .config import Config
#from .provider import ServiceProvider



from typing import Dict


class Config():
    config: Dict = {}

    def get(self, dotkey=None, config=None):
        # Recursive for dot notation
        if not dotkey:
            return self.config
        if config is None: config = self.config
        if "." in dotkey:
            key, rest = dotkey.split(".", 1)
            return self.get(rest, config[key])
        else:
            return config[dotkey]

    def set(self, dotkey, value, config=None):
        # Recursive for dot notation
        # Remember objects are byRef, so changing config also changes self.config
        if config is None: config = self.config
        if "." in dotkey:
            key, rest = dotkey.split(".", 1)
            if key not in config:
                config[key] = {}
            return self.set(rest, value, config[key])
        else:
            config[dotkey] = value
            return value

        #self.config[dotkey] = value
        #return value

    def __call__(self, dotkey=None):
        return self.get(dotkey)



import os
import sys
from typing import List

from ..http import Server
from ..support.config import Config
from ..support.helpers import import_module
from .cli import cli
from .version import version


class Application:

    # Public attributes
    version: str = version
    http: Server
    cli: cli
    config: Config
    providers: List = []


    perfs: List = []
    registered: bool = False
    booted: bool = False


    is_console: bool = False
    base_path: str = None
    name: str = None
    debug: bool = False

    def __init__(self, config: Config):
        self.config = config
        self.providers = []

    def bootstrap(self, name: str, base_path: str, is_console: bool):
        # Don't bootstrap multiple times
        if self.booted: return self

        # App name and base_path
        self.name = name
        self.base_path = base_path

        # Detect if running in console (to register commands)
        # Ensure console is False even when running ./uvicore http serve
        self.is_console = is_console
        if "'http', 'serve'" in str(sys.argv):
            self.is_console = False

        # Always set the cli instance, though commands won't be added if not is_console
        self.cli = cli

        # Add main app config
        app_config = import_module(self.name + '.config.app.app')[0]
        self.config.set('app', app_config)

        # Detect debug flag from main app config
        self.debug = app_config['debug']

        # Perf
        self.perf('|-foundation.application.bootstrap()')
        self.perf('|--is_console: ' + str(self.is_console))

        # Create our HTTP instance
        if not self.is_console:
            self.perf('|--firing up HTTP server')
            self.http = Server()

        # Build recursive providers graph
        self.build_provider_graph(self.name)
        #self.perf('--' + str(self.providers))

        # Register all providers
        self.register_providers()
        self.registered = True

        # Boot all providers
        # Not sure there will be a point yet?
        self.boot_providers()
        self.booted = True

        # Return application
        return self

    def register_providers(self):
        self.perf('|--registering providers')
        for (app, module) in self.providers:
            path = app + '.' + module
            self.perf('|---' + path)
            provider = import_module(path)[0](self)
            provider.register()

    def boot_providers(self):
        self.perf('|--booting providers')
        for (app, module) in self.providers:
            path = app + '.' + module
            self.perf('|---' + path)
            provider = import_module(path)[0](self)
            provider.boot()

    def build_provider_graph(self, app, module=None):
        if module:
            self.providers.append((app, module))
        app_config = import_module(app + '.config.app.app')[0]
        providers = app_config['providers']
        for (app, module) in providers:
            #app, module = provider
            #path = app + '.' + module
            if (app, module) not in self.providers:
                self.build_provider_graph(app, module)

    def perf(self, item):
        if self.debug:
            self.perfs.append(item)
            print(item)



import click, typer
from click_help_colors import HelpColorsGroup, HelpColorsCommand


group_kargs = {
    #'context_settings': dict(help_option_names=['-h', '--help']),
    'cls': HelpColorsGroup,
    'help_headers_color': 'yellow',
    'help_options_color': 'green',
}

cmd_kargs = {
    #'context_settings': dict(help_option_names=['-h', '--help']),
    'cls': HelpColorsCommand,
    'help_headers_color': 'yellow',
    'help_options_color': 'green',
}




# App and providers DO run if APP or LIBRARY mode
# Because even if module, it is a dependency graph
# So provider IMPORTER needs to not import things twice

# Providers should be a recursive dependency graph

app = {
    'name': 'uvicore.foundation',
    'version': '1.0.1',

    # Application Service Providers
    'providers': [
        # mRcore Framework
        #'mrcore.support',
    ]
}


# Package config with used as a package library
package = {

}

# class Config:
#     test = 'test-class'



from uvicore.support.click import click, group_kargs

from .version import version


@click.group(**group_kargs, help=f"""
    \b
    Uvicore {version}
    A Fast Async ASGI Python Framework for CLI, Web and API
    Copyright (c) 2020 Matthew Reschke
    License http://mreschke.com/license/mit
""")
@click.version_option(version=version, prog_name='Uvv', flag_value='--d')
def cli():
    pass




# cli = typer.Typer(
#     cls=HelpColorsGroup,
#     help_headers_color='yellow',
#     help_options_color='green',
#     help="""\b
# Uvicore Python Framework
# Copyright (c) 2020 Matthew Reschke
# License MIT
# """)

# import click


# #@click.group(**group_kargs)
# @click.group()
# def cli():
#     """
#     \b
#     mRcore Python Framework for CLI, Web and API
#     Copyright (c) 2020 Matthew Reschke
#     License http://mreschke.com/license/mit
#     """
#     pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uvicore',
 'uvicore.db',
 'uvicore.foundation',
 'uvicore.foundation.commands',
 'uvicore.foundation.config',
 'uvicore.http',
 'uvicore.support']

package_data = \
{'': ['*']}

install_requires = \
['alembic>=1.4.2,<2.0.0',
 'click-help-colors>=0.8,<0.9',
 'databases>=0.3.2,<0.4.0',
 'fastapi>=0.57.0,<0.58.0',
 'gunicorn>=20.0.4,<21.0.0',
 'typer>=0.2.1,<0.3.0',
 'uvicorn>=0.11.5,<0.12.0']

setup_kwargs = {
    'name': 'uvicore',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Matthew Reschke',
    'author_email': 'mail@mreschke.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from uvicore.support.provider import ServiceProvider


class Foundation(ServiceProvider):

    def register(self):

        # Register configs
        self.configs([
            ('foundation', 'uvicore.foundation.config.foundation.config')
        ])

        # # Register root level commands
        # self.commands(
        #     name='root',
        #     #help='xx',
        #     commands=[
        #         ('version', 'uvicore.foundation.commands.version.cli'),
        #     ]
        # )

        # Register HTTP Serve commands
        # force = False
        # if "'http', 'serve'" in str(sys.argv):
        #     force = True
        self.commands(
            name='http',
            help='Uvicore HTTP Commands',
            commands=[
                ('serve', 'uvicore.foundation.commands.serve.cli'),
            ],
            force=True  # Runs even if is_console = False
        )


    def boot(self):
        pass
